# bluemira is an integrated inter-disciplinary design tool for future fusion
# reactors. It incorporates several modules, some of which rely on other
# codes, to carry out a range of typical conceptual fusion reactor design
# activities.
#
# Copyright (C) 2021-2022 M. Coleman, J. Cook, F. Franza, I.A. Maione, S. McIntosh,
#                         J. Morris, D. Short
#
# bluemira is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# bluemira is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with bluemira; if not, see <https://www.gnu.org/licenses/>.

"""
Module containing the base Component class.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Sized,
    Tuple,
    Union,
)

import anytree
from anytree import NodeMixin, RenderTree

from bluemira.base.error import ComponentError
from bluemira.display.displayer import DisplayableCAD
from bluemira.display.plotter import Plottable

if TYPE_CHECKING:
    from bluemira.geometry.base import BluemiraGeo


class Component(NodeMixin, Plottable, DisplayableCAD):
    """
    The Component is the fundamental building block for a bluemira reactor design. It
    encodes the way that the corresponding part of the reactor will be built, along with
    any other derived properties that relate to that component.

    Components define a tree structure, based on the parent and children properties. This
    allows the nodes on that tree to be passed around within bluemira so that
    operations can be performed on the child branches of that structure.

    For example, a reactor design including just a TFCoilSystem may look as below:

    .. digraph:: base_component_tree

      "FusionPowerPlant" -> "TFCoilSystem" -> {"TFWindingPack" "TFCasing"}

    A Component cannot be used directly - only subclasses should be instantiated.
    """

    name: str

    def __init__(
        self,
        name: str,
        parent: Optional[Component] = None,
        children: Optional[List[Component]] = None,
    ):
        super().__init__()
        self.name = name

        if parent is not None and name in (ch.name for ch in parent.children):
            raise ComponentError(f"Component {name} is already a child of {parent}")

        self.parent = parent

        if children:
            self.children = children

    def __repr__(self) -> str:
        """
        The string representation of the instance
        """
        return self.name + " (" + self.__class__.__name__ + ")"

    def tree(self) -> str:
        """
        Get the tree of descendants of this instance.
        """
        return str(RenderTree(self))

    def get_component(
        self, name: str, first: bool = True, full_tree: bool = False
    ) -> Union["Component", List[Component], None]:
        """
        Find the components with the specified name.

        Parameters
        ----------
        name: str
            The name of the component to search for.
        first: bool
            If True, only the first element is returned, by default True.
        full_tree: bool
            If True, searches the tree from the root, else searches from this node, by
            default False.

        Returns
        -------
        found_components: Union[Component, List[Component]]
            The first component of the search if first is True, else all components
            matching the search.

        Notes
        -----
            This function is just a wrapper of the anytree.search.findall
            function.
        """
        return self._get_thing(
            lambda n: anytree.search._filter_by_name(n, "name", name), first, full_tree
        )

    def get_component_properties(
        self, properties: Union[Sized[str], str], first=True, full_tree=False
    ) -> Union[Tuple[List[Any]], List[Any], Any]:
        """
        Get properties from a component

        Parameters
        ----------
        name: str
            The name of the component to search for.
        first: bool
            If True, only the first element is returned, by default True.
        full_tree: bool
            If True, searches the tree from the root, else searches from this node, by
            default False.

        Returns
        -------
        properties: Union[Tuple[List[Any]], List[Any], Any]
            If multiple properties specified returns a tuple of the list of properties,
            otherwise returns a list of the property.
            If only one node has the property returns the value(s).

        Notes
        -----
            This function is just a wrapper of the anytree.search.findall or find
            functions.
        """
        if isinstance(properties, str):
            properties = [properties]

        def filter_(node, properties):
            return all(hasattr(node, prop) for prop in properties)

        found_nodes = self._get_thing(lambda n: filter_(n, properties), first, full_tree)

        if found_nodes is None:
            return tuple([] for _ in properties)

        if not isinstance(found_nodes, Iterable):
            if len(properties) == 1:
                return getattr(found_nodes, properties[0])
            return [getattr(found_nodes, prop) for prop in properties]
        else:
            # Collect values by property instead of by node
            node_properties = [
                [getattr(node, prop) for prop in properties] for node in found_nodes
            ]
            return tuple(map(list, zip(*node_properties)))

    def _get_thing(self, filter_: Callable, first: bool, full_tree: bool):
        found_nodes = anytree.search.findall(
            self.root if full_tree else self, filter_=filter_
        )
        if found_nodes in (None, ()):
            return None
        if first and isinstance(found_nodes, Iterable):
            found_nodes = found_nodes[0]
        return found_nodes

    def add_child(self, child: Component) -> Component:
        """
        Add a single child to this node

        Parameters
        ----------
        child: Component
            The child to be added

        Returns
        -------
        self: Component
            This component.
        """
        # TODO: Support merge_trees here too.
        if child in self.children or child.name in (ch.name for ch in self.children):
            raise ComponentError(f"Component {child} is already a child of {self}")
        self.children = list(self.children) + [child]

        return self

    def add_children(self, children: List[Component], merge_trees=False):
        """
        Add multiple children to this node

        Parameters
        ----------
        children: List[Component]
            The children to be added

        Returns
        -------
        self: Component
            This component.
        """
        if not isinstance(children, list) or len(children) == 0:
            child = children[0] if isinstance(children, list) else children
            return self.add_child(child)

        duplicates = []
        child: Component
        for idx, child in reversed(list(enumerate(children))):
            existing = self.get_component(child.name)
            if existing is not None:
                if merge_trees:
                    existing.children = list(existing.children) + list(child.children)
                    children.pop(idx)
                else:
                    duplicates += [child]
        if duplicates != []:
            raise ComponentError(
                f"Components {duplicates} are already children of {self}"
            )
        self.children = list(self.children) + children

        return self

    def prune_child(self, name: str):
        """
        Remove the child with the given name, and all its children.
        """
        found_component = anytree.search.find_by_attr(self, name)
        if found_component:
            # Method of deleting a node suggested by library author
            # https://github.com/c0fec0de/anytree/issues/152
            found_component.parent = None


class PhysicalComponent(Component):
    """
    A physical component. It includes shape and materials.
    """

    def __init__(
        self,
        name: str,
        shape: BluemiraGeo,
        material: Any = None,
        parent: Component = None,
        children: Component = None,
    ):
        super().__init__(name, parent, children)
        self.shape = shape
        self.material = material

    @property
    def shape(self) -> BluemiraGeo:
        """
        The geometric shape of the Component.
        """
        return self._shape

    @shape.setter
    def shape(self, value: BluemiraGeo):
        self._shape = value

    @property
    def material(self):
        """
        The material that the Component is built from.
        """
        return self._material

    @material.setter
    def material(self, value):
        self._material = value


class MagneticComponent(PhysicalComponent):
    """
    A magnetic component. It includes a shape, a material, and a source conductor.
    """

    def __init__(
        self,
        name: str,
        shape: BluemiraGeo,
        material: Any = None,
        conductor: Any = None,
        parent: Component = None,
        children: Component = None,
    ):
        super().__init__(name, shape, material, parent, children)
        self.conductor = conductor

    @property
    def conductor(self):
        """
        The conductor used by current-carrying filaments.
        """
        return self._conductor

    @conductor.setter
    def conductor(self, value):
        self._conductor = value


def get_properties_from_components(
    comps: Union[Component, Iterable[Component]], properties: Union[str, Sized[str]]
):
    """
    Get properties from Components

    Parameters
    ----------
    comps: Union[Component, Iterable[Component]])
        A component or list of components
    properties: Union[str, Sized[str]]
        properties to collect

    Returns
    -------
    properties: Union[Tuple[List[Any]], List[Any], Any]
        If multiple properties specified returns a tuple of the list of properties,
        otherwise returns a list of the property.
        If only one node has the property returns the value(s).

    """
    if isinstance(properties, str):
        properties = [properties]

    property_lists = tuple([] for _ in properties)

    if not isinstance(comps, Iterable):
        comps = [comps]

    for comp in comps:

        props = comp.get_component_properties(properties, first=False)
        if not isinstance(props, tuple):
            props = tuple(props)
        for i, prop in enumerate(props):
            property_lists[i].extend(prop)

    if len(property_lists[0]) == 1:
        property_lists = [p[0] for p in property_lists]

    if len(property_lists) == 1:
        property_lists = property_lists[0]

    return property_lists
